@@ -1,13 +1,19 @@
 /*
  * Copyright (c) 2007 Vreixo Formoso
+ * Copyright (c) 2010 Thomas Schmitt
  *
  * This file is part of the libisofs project; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License version 2 as
- * published by the Free Software Foundation. See COPYING file for details.
+ * modify it under the terms of the GNU General Public License version 2 
+ * or later as published by the Free Software Foundation. 
+ * See COPYING file for details.
  */
 
+#ifdef HAVE_CONFIG_H
+#include "../config.h"
+#endif
+
+#include "libisofs.h"
 #include "eltorito.h"
-#include "stream.h"
 #include "fsource.h"
 #include "filesrc.h"
 #include "image.h"
@@ -16,9 +22,10 @@
 
 #include <stdlib.h>
 #include <string.h>
+#include <stdio.h>
 
 /**
- * This table should be written with accuracy values at offset
+ * This table should be written with the actual values at offset
  * 8 of boot image, when used ISOLINUX boot loader
  */
 struct boot_info_table {
@@ -54,6 +61,19 @@ struct hard_disc_mbr {
     uint8_t sign2;
 };
 
+/* API */
+int el_torito_set_boot_platform_id(ElToritoBootImage *bootimg, uint8_t id)
+{
+    bootimg->platform_id = id;
+    return 1;
+}
+
+/* API */
+int el_torito_get_boot_platform_id(ElToritoBootImage *bootimg)
+{
+    return bootimg->platform_id;
+}
+
 /**
  * Sets the load segment for the initial boot image. This is only for
  * no emulation boot images, and is a NOP for other image types.
@@ -65,6 +85,14 @@ void el_torito_set_load_seg(ElToritoBootImage *bootimg, short segment)
     bootimg->load_seg = segment;
 }
 
+/* API */
+int el_torito_get_load_seg(ElToritoBootImage *bootimg)
+{
+   if (bootimg->load_seg < 0)
+       return 0xffff - bootimg->load_seg;
+   return bootimg->load_seg;
+}
+ 
 /**
  * Sets the number of sectors (512b) to be load at load segment during
  * the initial boot procedure. This is only for no emulation boot images,
@@ -77,6 +105,14 @@ void el_torito_set_load_size(ElToritoBootImage *bootimg, short sectors)
     bootimg->load_size = sectors;
 }
 
+/* API */
+int el_torito_get_load_size(ElToritoBootImage *bootimg)
+{
+   if (bootimg->load_size < 0)
+       return 0xffff - bootimg->load_size;
+   return bootimg->load_size;
+}
+
 /**
  * Marks the specified boot image as not bootable
  */
@@ -85,8 +121,50 @@ void el_torito_set_no_bootable(ElToritoBootImage *bootimg)
     bootimg->bootable = 0;
 }
 
+/* API */
+int el_torito_get_bootable(ElToritoBootImage *bootimg)
+{   
+    return !!bootimg->bootable;
+}
+
+/* API */
+int el_torito_set_id_string(ElToritoBootImage *bootimg, uint8_t id_string[28])
+{
+    memcpy(bootimg->id_string, id_string, 28);
+    return 1;
+}
+
+/* API */
+int el_torito_get_id_string(ElToritoBootImage *bootimg, uint8_t id_string[28])
+{
+    
+    memcpy(id_string, bootimg->id_string, 28);
+    return 1;
+}
+
+/* API */
+int el_torito_set_selection_crit(ElToritoBootImage *bootimg, uint8_t crit[20])
+{
+    memcpy(bootimg->selection_crit, crit, 20);
+    return 1;
+}
+
+/* API */
+int el_torito_get_selection_crit(ElToritoBootImage *bootimg, uint8_t crit[20])
+{
+    
+    memcpy(crit, bootimg->selection_crit, 20);
+    return 1;
+}
+
+/* API */
+int el_torito_seems_boot_info_table(ElToritoBootImage *bootimg, int flag)
+{
+    return bootimg->seems_boot_info_table;
+}
+
 /**
- * Specifies that this image needs to be patched. This involves the writting
+ * Specifies that this image needs to be patched. This involves the writing
  * of a 56 bytes boot information table at offset 8 of the boot image file.
  * The original boot image file won't be modified.
  * This is needed for isolinux boot images.
@@ -105,7 +183,7 @@ void el_torito_patch_isolinux_image(ElToritoBootImage *bootimg)
  *        bitmask style flag. The following values are defined:
  *
  *        bit 0 -> 1 to path the image, 0 to not
- *                 Patching the image involves the writting of a 56 bytes
+ *                 Patching the image involves the writing of a 56 bytes
  *                 boot information table at offset 8 of the boot image file.
  *                 The original boot image file won't be modified. This is needed
  *                 to allow isolinux images to be bootable.
@@ -119,35 +197,40 @@ void el_torito_patch_isolinux_image(ElToritoBootImage *bootimg)
  */
 int el_torito_set_isolinux_options(ElToritoBootImage *bootimg, int options, int flag)
 {
-    bootimg->isolinux_options = (options & 0x03);
+    bootimg->isolinux_options = (options & 0x01ff);
     return ISO_SUCCESS;
 }
 
-/* TODO getter for boot image properties should be exposed
- * useful when reading discs */
-int el_torito_get_boot_media_type(const ElToritoBootImage *bootimg)
+/* API */
+int el_torito_get_isolinux_options(ElToritoBootImage *bootimg, int flag)
+{
+    return bootimg->isolinux_options & 0x01ff;
+}
+
+/* API */
+int el_torito_get_boot_media_type(ElToritoBootImage *bootimg,
+                                  enum eltorito_boot_media_type *media_type)
 {
     if (bootimg) {
         switch (bootimg->type) {
         case 1:
         case 2:
         case 3:
-            return ELTORITO_FLOPPY_EMUL;
-            break;
+            *media_type = ELTORITO_FLOPPY_EMUL;
+            return 1;
         case 4:
-            return ELTORITO_HARD_DISC_EMUL;
-            break;
+            *media_type = ELTORITO_HARD_DISC_EMUL;
+            return 1;
         case 0:
-            return ELTORITO_NO_EMUL;
-            break;
+            *media_type = ELTORITO_NO_EMUL;
+            return 1;
         default:
             /* should never happen */
             return ISO_ASSERT_FAILURE;
             break;
         }
-    } else {
-        return ISO_WRONG_ARG_VALUE;
     }
+    return ISO_WRONG_ARG_VALUE;
 }
 
 static
@@ -156,6 +239,7 @@ int iso_tree_add_boot_node(IsoDir *parent, const char *name, IsoBoot **boot)
     IsoBoot *node;
     IsoNode **pos;
     time_t now;
+    int ret;
 
     if (parent == NULL || name == NULL || boot == NULL) {
         return ISO_NULL_POINTER;
@@ -165,9 +249,9 @@ int iso_tree_add_boot_node(IsoDir *parent, const char *name, IsoBoot **boot)
     }
 
     /* check if the name is valid */
-    if (!iso_node_is_valid_name(name)) {
-        return ISO_WRONG_ARG_VALUE;
-    }
+    ret = iso_node_is_valid_name(name);
+    if (ret < 0)
+        return ret;
 
     /* find place where to insert */
     pos = &(parent->children);
@@ -191,6 +275,9 @@ int iso_tree_add_boot_node(IsoDir *parent, const char *name, IsoBoot **boot)
         free(node);
         return ISO_OUT_OF_MEM;
     }
+    node->lba = 0;
+    node->size = 0;
+    node->content = NULL;
 
     /* atributes from parent */
     node->node.mode = S_IFREG | (parent->node.mode & 0444);
@@ -226,6 +313,7 @@ int create_image(IsoImage *image, const char *image_path,
     int boot_media_type = 0;
     int load_sectors = 0; /* number of sector to load */
     unsigned char partition_type = 0;
+    off_t size;
     IsoNode *imgfile;
     IsoStream *stream;
 
@@ -234,6 +322,9 @@ int create_image(IsoImage *image, const char *image_path,
         return ret;
     }
     if (ret == 0) {
+        iso_msg_submit(image->id, ISO_NODE_DOESNT_EXIST, 0,
+                       "El Torito boot image file missing in ISO image: '%s'",
+                       image_path);
         return ISO_NODE_DOESNT_EXIST;
     }
 
@@ -248,9 +339,16 @@ int create_image(IsoImage *image, const char *image_path,
         return ISO_BOOT_IMAGE_NOT_VALID;
     }
 
+    size = iso_stream_get_size(stream);
+    if (size <= 0) {
+        iso_msg_submit(image->id, ISO_BOOT_IMAGE_NOT_VALID, 0,
+                       "Boot image file is empty");
+        return ISO_BOOT_IMAGE_NOT_VALID;
+    }
+
     switch (type) {
     case ELTORITO_FLOPPY_EMUL:
-        switch (iso_stream_get_size(stream)) {
+        switch (size) {
         case 1200 * 1024:
             boot_media_type = 1; /* 1.2 meg diskette */
             break;
@@ -289,14 +387,14 @@ int create_image(IsoImage *image, const char *image_path,
         if (ret != sizeof(mbr)) {
             iso_msg_submit(image->id, ISO_BOOT_IMAGE_NOT_VALID, 0,
                           "Can't read MBR from image file.");
-            return ret < 0 ? ret : ISO_FILE_READ_ERROR;
+            return ret < 0 ? ret : (int) ISO_FILE_READ_ERROR;
         }
 
         /* check valid MBR signature */
         if ( mbr.sign1 != 0x55 || mbr.sign2 != 0xAA ) {
             iso_msg_submit(image->id, ISO_BOOT_IMAGE_NOT_VALID, 0,
                           "Invalid MBR. Wrong signature.");
-            return ISO_BOOT_IMAGE_NOT_VALID;
+            return (int) ISO_BOOT_IMAGE_NOT_VALID;
         }
 
         /* ensure single partition */
@@ -332,10 +430,15 @@ int create_image(IsoImage *image, const char *image_path,
     boot->image = (IsoFile*)imgfile;
     iso_node_ref(imgfile); /* get our ref */
     boot->bootable = 1;
+    boot->seems_boot_info_table = 0;
+    boot->isolinux_options = 0;
     boot->type = boot_media_type;
-    boot->load_size = load_sectors;
     boot->partition_type = partition_type;
-
+    boot->load_seg = 0;
+    boot->load_size = load_sectors;
+    boot->platform_id = 0; /* 80x86 */
+    memset(boot->id_string, 0, sizeof(boot->id_string));
+    memset(boot->selection_crit, 0, sizeof(boot->selection_crit));
     if (bootimg) {
         *bootimg = boot;
     }
@@ -348,7 +451,7 @@ int iso_image_set_boot_image(IsoImage *image, const char *image_path,
                              const char *catalog_path,
                              ElToritoBootImage **boot)
 {
-    int ret;
+    int ret, i;
     struct el_torito_boot_catalog *catalog;
     ElToritoBootImage *boot_image= NULL;
     IsoBoot *cat_node= NULL;
@@ -383,8 +486,11 @@ int iso_image_set_boot_image(IsoImage *image, const char *image_path,
             catname[0] = '\0';
             ret = iso_tree_path_to_node(image, catdir, &p);
             if (ret <= 0) {
+                iso_msg_submit(image->id, ISO_NODE_DOESNT_EXIST, 0,
+         "Cannot find directory for El Torito boot catalog in ISO image: '%s'",
+                               catdir);
                 free(catdir);
-                return ret < 0 ? ret : ISO_NODE_DOESNT_EXIST;
+                return ret < 0 ? ret : (int) ISO_NODE_DOESNT_EXIST;
             }
             if (p->type != LIBISO_DIR) {
                 free(catdir);
@@ -407,13 +513,17 @@ int iso_image_set_boot_image(IsoImage *image, const char *image_path,
     }
 
     /* creates the catalog with the given image */
-    catalog = malloc(sizeof(struct el_torito_boot_catalog));
+    catalog = calloc(1, sizeof(struct el_torito_boot_catalog));
     if (catalog == NULL) {
         ret = ISO_OUT_OF_MEM;
         goto boot_image_cleanup;
     }
-    catalog->image = boot_image;
+    catalog->num_bootimages = 1;
+    catalog->bootimages[0] = boot_image;
+    for (i = 1; i < Libisofs_max_boot_imageS; i++)
+        catalog->bootimages[i] = NULL;
     catalog->node = cat_node;
+    catalog->sort_weight = 1000;                            /* slightly high */
     iso_node_ref((IsoNode*)cat_node);
     image->bootcat = catalog;
 
@@ -436,7 +546,7 @@ boot_image_cleanup:;
 }
 
 /**
- * Get El-Torito boot image of an ISO image, if any.
+ * Get the boot catalog and the El-Torito default boot image of an ISO image.
  *
  * This can be useful, for example, to check if a volume read from a previous
  * session or an existing image is bootable. It can also be useful to get
@@ -480,10 +590,10 @@ int iso_image_get_boot_image(IsoImage *image, ElToritoBootImage **boot,
 
     /* ok, image is bootable */
     if (boot) {
-        *boot = image->bootcat->image;
+        *boot = image->bootcat->bootimages[0];
     }
     if (imgnode) {
-        *imgnode = image->bootcat->image->image;
+        *imgnode = image->bootcat->bootimages[0]->image;
     }
     if (catnode) {
         *catnode = image->bootcat->node;
@@ -491,6 +601,64 @@ int iso_image_get_boot_image(IsoImage *image, ElToritoBootImage **boot,
     return ISO_SUCCESS;
 }
 
+int iso_image_get_bootcat(IsoImage *image, IsoBoot **catnode, uint32_t *lba,
+                          char **content, off_t *size)
+{
+    IsoBoot *bootcat;
+
+    *catnode = NULL;
+    *lba = 0;
+    *content = NULL;
+    *size = 0;
+    bootcat = image->bootcat->node;
+    if (bootcat == NULL)
+        return 0;
+    *catnode = bootcat;
+    *lba = bootcat->lba;
+    *size = bootcat->size;
+    if (bootcat->size > 0 && bootcat->content != NULL) {
+        *content = calloc(1, bootcat->size);
+        if (*content == NULL) 
+            return ISO_OUT_OF_MEM;
+        memcpy(*content, bootcat->content, bootcat->size);
+    }
+    return 1;
+}
+
+int iso_image_get_all_boot_imgs(IsoImage *image, int *num_boots,
+               ElToritoBootImage ***boots, IsoFile ***bootnodes, int flag)
+{
+    int i;
+    struct el_torito_boot_catalog *cat;
+
+    if (image == NULL)
+        return ISO_NULL_POINTER;
+    if (image->bootcat == NULL)
+        return 0;
+    cat = image->bootcat;
+    *num_boots = cat->num_bootimages;
+    *boots = NULL;
+    *bootnodes = NULL;
+    if (*num_boots <= 0)
+        return 0;
+    *boots = calloc(*num_boots, sizeof(ElToritoBootImage *));
+    *bootnodes = calloc(*num_boots, sizeof(IsoFile *));
+    if(*boots == NULL || *bootnodes == NULL) {
+        if (*boots != NULL)
+            free(*boots);
+        if (*bootnodes != NULL)
+            free(*bootnodes);
+        *boots = NULL;
+        *bootnodes = NULL;
+        return ISO_OUT_OF_MEM;
+    }
+    for (i = 0; i < *num_boots; i++) {
+        (*boots)[i] = cat->bootimages[i];
+        (*bootnodes)[i] = image->bootcat->bootimages[i]->image;
+    }
+    return 1;
+}
+
 /**
  * Removes the El-Torito bootable image.
  *
@@ -505,28 +673,82 @@ void iso_image_remove_boot_image(IsoImage *image)
         return;
 
     /*
-     * remove catalog node from its parent
-     * (the reference will be disposed next)
+     * remove catalog node from its parent and dispose it
+     * (another reference is with the catalog)
      */
-    iso_node_take((IsoNode*)image->bootcat->node);
+    if (iso_node_get_parent((IsoNode*) image->bootcat->node) != NULL) {
+        iso_node_take((IsoNode*) image->bootcat->node);
+        iso_node_unref((IsoNode*) image->bootcat->node);
+    }
 
     /* free boot catalog and image, including references to nodes */
     el_torito_boot_catalog_free(image->bootcat);
     image->bootcat = NULL;
 }
 
+/* API */
+int iso_image_add_boot_image(IsoImage *image, const char *image_path,
+                             enum eltorito_boot_media_type type, int flag,
+                             ElToritoBootImage **boot)
+{
+    int ret;
+    struct el_torito_boot_catalog *catalog = image->bootcat;
+    ElToritoBootImage *boot_img;
+
+    if(catalog == NULL)
+      return ISO_BOOT_NO_CATALOG;
+    if (catalog->num_bootimages >= Libisofs_max_boot_imageS)
+        return ISO_BOOT_IMAGE_OVERFLOW;
+    ret = create_image(image, image_path, type, &boot_img);
+    if (ret < 0) 
+        return ret;
+    catalog->bootimages[catalog->num_bootimages] = boot_img;
+    catalog->num_bootimages++;
+    if (boot != NULL)
+        *boot = boot_img;
+    return 1;
+}
+
+/* API */
+int iso_image_set_boot_catalog_weight(IsoImage *image, int sort_weight)
+{
+    if (image->bootcat == NULL)
+        return 0;
+    image->bootcat->sort_weight = sort_weight;
+    return 1;
+}
+
+/* API */
+int iso_image_set_boot_catalog_hidden(IsoImage *image, int hide_attrs)
+{
+    if (image->bootcat == NULL)
+        return 0;
+    if (image->bootcat->node == NULL)
+        return 0;
+    iso_node_set_hidden((IsoNode *) image->bootcat->node, hide_attrs);
+    return 1;
+}
+
+
 void el_torito_boot_catalog_free(struct el_torito_boot_catalog *cat)
 {
     struct el_torito_boot_image *image;
+    int i;
 
     if (cat == NULL) {
         return;
     }
 
-    image = cat->image;
-    iso_node_unref((IsoNode*)image->image);
-    free(image);
-    iso_node_unref((IsoNode*)cat->node);
+    for (i = 0; i < Libisofs_max_boot_imageS; i++) {
+        image = cat->bootimages[i];
+        if (image == NULL)
+    continue;
+        if ((IsoNode*)image->image != NULL)
+            iso_node_unref((IsoNode*)image->image);
+        free(image);
+    }
+    if ((IsoNode*)cat->node != NULL)
+        iso_node_unref((IsoNode*)cat->node);
     free(cat);
 }
 
@@ -537,11 +759,12 @@ struct catalog_stream
 {
     Ecma119Image *target;
     uint8_t buffer[BLOCK_SIZE];
-    int offset; /* -1 if stream is not openned */
+    int offset; /* -1 if stream is not opened */
 };
 
 static void
-write_validation_entry(uint8_t *buf)
+write_validation_entry(uint8_t *buf, uint8_t platform_id,
+                       uint8_t id_string[24])
 {
     size_t i;
     int checksum;
@@ -549,10 +772,10 @@ write_validation_entry(uint8_t *buf)
     struct el_torito_validation_entry *ve =
         (struct el_torito_validation_entry*)buf;
     ve->header_id[0] = 1;
-    ve->platform_id[0] = 0; /* 0: 80x86, 1: PowerPC, 2: Mac */
+    ve->platform_id[0] = platform_id;
+    memcpy(ve->id_string, id_string, sizeof(ve->id_string));
     ve->key_byte1[0] = 0x55;
     ve->key_byte2[0] = 0xAA;
-
     /* calculate the checksum, to ensure sum of all words is 0 */
     checksum = 0;
     for (i = 0; i < sizeof(struct el_torito_validation_entry); i += 2) {
@@ -561,35 +784,63 @@ write_validation_entry(uint8_t *buf)
     iso_lsb(ve->checksum, checksum, 2);
 }
 
+static void
+write_section_header(uint8_t *buf, Ecma119Image *t, int idx, int num_entries)
+{
+    char *id_string;
+
+    struct el_torito_section_header *e =
+        (struct el_torito_section_header *) buf;
+
+    /* 0x90 = more section headers follow , 0x91 = final section */
+    e->header_indicator[0] = 0x90 + (idx == t->catalog->num_bootimages - 1);
+    e->platform_id[0] = t->catalog->bootimages[idx]->platform_id;
+    e->num_entries[0] = num_entries & 0xff;
+    e->num_entries[1] = (num_entries >> 8) & 0xff;;
+    id_string = (char *) e->id_string;
+    memcpy(id_string,  t->catalog->bootimages[idx]->id_string,
+           sizeof(e->id_string));
+}
+
 /**
  * Write one section entry.
- * Currently this is used only for default image (the only supported just now)
+ * Usable for the Default Entry
+ * and for Section Entries with Selection criteria type == 0
  */
 static void
-write_section_entry(uint8_t *buf, Ecma119Image *t)
+write_section_entry(uint8_t *buf, Ecma119Image *t, int idx)
 {
     struct el_torito_boot_image *img;
     struct el_torito_section_entry *se =
         (struct el_torito_section_entry*)buf;
 
-    img = t->catalog->image;
+    img = t->catalog->bootimages[idx];
 
     se->boot_indicator[0] = img->bootable ? 0x88 : 0x00;
     se->boot_media_type[0] = img->type;
     iso_lsb(se->load_seg, img->load_seg, 2);
     se->system_type[0] = img->partition_type;
     iso_lsb(se->sec_count, img->load_size, 2);
-    iso_lsb(se->block, t->bootimg->sections[0].block, 4);
+    iso_lsb(se->block, t->bootsrc[idx]->sections[0].block, 4);
+    se->selec_criteria[0] = img->selection_crit[0];
+    memcpy(se->vendor_sc, img->selection_crit + 1, 19);
 }
 
 static
 int catalog_open(IsoStream *stream)
 {
+    int i, j, k, num_entries;
     struct catalog_stream *data;
+    uint8_t *wpt;
+    struct el_torito_boot_catalog *cat;
+    struct el_torito_boot_image **boots;
+
     if (stream == NULL) {
         return ISO_NULL_POINTER;
     }
     data = stream->data;
+    cat = data->target->catalog;
+    boots = cat->bootimages;
 
     if (data->offset != -1) {
         return ISO_FILE_ALREADY_OPENED;
@@ -598,11 +849,36 @@ int catalog_open(IsoStream *stream)
     memset(data->buffer, 0, BLOCK_SIZE);
 
     /* fill the buffer with the catalog contents */
-    write_validation_entry(data->buffer);
+    write_validation_entry(data->buffer,
+                           boots[0]->platform_id, boots[0]->id_string);
 
-    /* write default entry */
-    write_section_entry(data->buffer + 32, data->target);
+    /* write default entry = first boot image */
+    write_section_entry(data->buffer + 32, data->target, 0);
 
+    /* IMPORTANT: The maximum number of boot images must fit into BLOCK_SIZE */
+    wpt = data->buffer + 64;
+    for (i = 1; i < cat->num_bootimages; ) {
+        /* Look ahead and put images of same platform_id and id_string
+           into the same section */
+        for (j = i + 1; j < cat->num_bootimages; j++) {
+             if (boots[i]->platform_id != boots[j]->platform_id)
+        break;
+             for (k = 0; k < (int) sizeof(boots[i]->id_string); k++)
+                 if (boots[i]->id_string[k] != boots[j]->id_string[k])
+             break;
+             if (k < (int) sizeof(boots[i]->id_string))
+        break;
+        }
+        num_entries = j - i;
+
+        write_section_header(wpt, data->target, i, num_entries);
+        wpt += 32;
+        for (j = 0; j < num_entries; j++) {
+            write_section_entry(wpt,  data->target, i);
+            wpt += 32;
+            i++;
+        }
+    }
     data->offset = 0;
     return ISO_SUCCESS;
 }
@@ -646,7 +922,7 @@ int catalog_read(IsoStream *stream, void *buf, size_t count)
         return ISO_FILE_NOT_OPENED;
     }
 
-    len = MIN(count, BLOCK_SIZE - data->offset);
+    len = MIN(count, (size_t) (BLOCK_SIZE - data->offset));
     memcpy(buf, data->buffer + data->offset, len);
     return len;
 }
@@ -686,7 +962,11 @@ IsoStreamIface catalog_stream_class = {
     catalog_read,
     catalog_is_repeatable,
     catalog_get_id,
-    catalog_free
+    catalog_free,
+    NULL,
+    NULL,
+    NULL,
+    NULL
 };
 
 /**
@@ -702,12 +982,12 @@ int catalog_stream_new(Ecma119Image *target, IsoStream **stream)
         return ISO_NULL_POINTER;
     }
 
-    str = malloc(sizeof(IsoStream));
+    str = calloc(1, sizeof(IsoStream));
     if (str == NULL) {
         return ISO_OUT_OF_MEM;
     }
-    data = malloc(sizeof(struct catalog_stream));
-    if (str == NULL) {
+    data = calloc(1, sizeof(struct catalog_stream));
+    if (data == NULL) {
         free(str);
         return ISO_OUT_OF_MEM;
     }
@@ -740,7 +1020,7 @@ int el_torito_catalog_file_src_create(Ecma119Image *target, IsoFileSrc **src)
         return ISO_SUCCESS;
     }
 
-    file = malloc(sizeof(IsoFileSrc));
+    file = calloc(1, sizeof(IsoFileSrc));
     if (file == NULL) {
         return ISO_OUT_OF_MEM;
     }
@@ -752,10 +1032,11 @@ int el_torito_catalog_file_src_create(Ecma119Image *target, IsoFileSrc **src)
     }
 
     /* fill fields */
-    file->prev_img = 0; /* TODO allow copy of old img catalog???? */
+    file->no_write = 0; /* TODO allow copy of old img catalog???? */
+    file->checksum_index = 0;
     file->nsections = 1;
     file->sections = calloc(1, sizeof(struct iso_file_section));
-    file->sort_weight = 1000; /* slightly high */
+    file->sort_weight = target->catalog->sort_weight;
     file->stream = stream;
 
     ret = iso_file_src_add(target, file, src);
@@ -771,27 +1052,26 @@ int el_torito_catalog_file_src_create(Ecma119Image *target, IsoFileSrc **src)
 /******************* EL-TORITO WRITER *******************************/
 
 /**
- * Patch an isolinux boot image.
+ * Insert boot info table content into buf.
  *
  * @return
  *      1 on success, 0 error (but continue), < 0 error
  */
-static
-int patch_boot_image(uint8_t *buf, Ecma119Image *t, size_t imgsize)
+int make_boot_info_table(uint8_t *buf, uint32_t pvd_lba,
+                         uint32_t boot_lba, uint32_t imgsize)
 {
     struct boot_info_table *info;
     uint32_t checksum;
-    size_t offset;
+    uint32_t offset;
 
-    if (imgsize < 64) {
-        return iso_msg_submit(t->image->id, ISO_ISOLINUX_CANT_PATCH, 0,
-            "Isolinux image too small. We won't patch it.");
-    }
+    info = (struct boot_info_table *) (buf + 8);
+    if (imgsize < 64)
+        return ISO_ISOLINUX_CANT_PATCH;
 
     /* compute checksum, as the the sum of all 32 bit words in boot image
      * from offset 64 */
     checksum = 0;
-    offset = (size_t) 64;
+    offset = 64;
 
     while (offset <= imgsize - 4) {
         checksum += iso_read_lsb(buf + offset, 4);
@@ -806,55 +1086,75 @@ int patch_boot_image(uint8_t *buf, Ecma119Image *t, size_t imgsize)
         checksum += iso_read_lsb(buf + offset, imgsize - offset);
     }
 
-    /* patch boot info table */
-    info = (struct boot_info_table*)(buf + 8);
     /*memset(info, 0, sizeof(struct boot_info_table));*/
-    iso_lsb(info->bi_pvd, t->ms_block + 16, 4);
-    iso_lsb(info->bi_file, t->bootimg->sections[0].block, 4);
+    iso_lsb(info->bi_pvd, pvd_lba, 4);
+    iso_lsb(info->bi_file, boot_lba, 4);
     iso_lsb(info->bi_length, imgsize, 4);
     iso_lsb(info->bi_csum, checksum, 4);
+    memset(buf + 24, 0, 40);
     return ISO_SUCCESS;
 }
 
+/**
+ * Patch an isolinux boot image.
+ *
+ * @return
+ *      1 on success, 0 error (but continue), < 0 error
+ */
 static
-int eltorito_writer_compute_data_blocks(IsoImageWriter *writer)
+int patch_boot_info_table(uint8_t *buf, Ecma119Image *t,
+                               size_t imgsize, int idx)
 {
-    /*
-     * We have nothing to write, but if we need to patch an isolinux image,
-     * this is a good place to do so.
-     */
-    Ecma119Image *t;
     int ret;
 
-    if (writer == NULL) {
-        return ISO_NULL_POINTER;
+    if (imgsize < 64) {
+        return iso_msg_submit(t->image->id, ISO_ISOLINUX_CANT_PATCH, 0,
+            "Isolinux image too small. We won't patch it.");
     }
+    ret = make_boot_info_table(buf, t->ms_block + (uint32_t) 16,
+                               t->bootsrc[idx]->sections[0].block,
+                               (uint32_t) imgsize);
+    return ret;
+}
 
-    t = writer->target;
+/* Patch the boot images if indicated */
+int iso_patch_eltoritos(Ecma119Image *t)
+{
+    int ret, idx;
+    size_t size;
+    uint8_t *buf;
+    IsoStream *new = NULL;
+    IsoStream *original = NULL;
 
-    if (t->catalog->image->isolinux_options & 0x01) {
-        /* we need to patch the image */
-        size_t size;
-        uint8_t *buf;
-        IsoStream *new = NULL;
-        IsoStream *original = t->bootimg->stream;
+    if (t->catalog == NULL)
+        return ISO_SUCCESS;
+
+    for (idx = 0; idx < t->catalog->num_bootimages; idx++) {
+        if (!(t->catalog->bootimages[idx]->isolinux_options & 0x01))
+    continue;
+        original = t->bootsrc[idx]->stream;
         size = (size_t) iso_stream_get_size(original);
-        buf = malloc(size);
+
+        /* >>> BOOT ts B00428 :
+               check whether size is not too large for buffering */;
+
+        buf = calloc(1, size);
         if (buf == NULL) {
             return ISO_OUT_OF_MEM;
         }
         ret = iso_stream_open(original);
         if (ret < 0) {
+            free(buf);
             return ret;
         }
         ret = iso_stream_read(original, buf, size);
         iso_stream_close(original);
-        if (ret != size) {
-            return (ret < 0) ? ret : ISO_FILE_READ_ERROR;
+        if (ret != (int) size) {
+            return (ret < 0) ? ret : (int) ISO_FILE_READ_ERROR;
         }
 
         /* ok, patch the read buffer */
-        ret = patch_boot_image(buf, t, size);
+        ret = patch_boot_info_table(buf, t, size, idx);
         if (ret < 0) {
             return ret;
         }
@@ -865,12 +1165,22 @@ int eltorito_writer_compute_data_blocks(IsoImageWriter *writer)
         if (ret < 0) {
             return ret;
         }
-        t->bootimg->stream = new;
+        t->bootsrc[idx]->stream = new;
         iso_stream_unref(original);
     }
     return ISO_SUCCESS;
 }
 
+static
+int eltorito_writer_compute_data_blocks(IsoImageWriter *writer)
+{
+    /*
+     * We have nothing to write.
+     */
+    return ISO_SUCCESS;
+}
+
+
 /**
  * Write the Boot Record Volume Descriptor (ECMA-119, 8.2)
  */
@@ -878,7 +1188,6 @@ static
 int eltorito_writer_write_vol_desc(IsoImageWriter *writer)
 {
     Ecma119Image *t;
-    struct el_torito_boot_catalog *cat;
     struct ecma119_boot_rec_vol_desc vol;
 
     if (writer == NULL) {
@@ -886,8 +1195,6 @@ int eltorito_writer_write_vol_desc(IsoImageWriter *writer)
     }
 
     t = writer->target;
-    cat = t->catalog;
-
     iso_msg_debug(t->image->id, "Write El-Torito boot record");
 
     memset(&vol, 0, sizeof(struct ecma119_boot_rec_vol_desc));
@@ -895,8 +1202,8 @@ int eltorito_writer_write_vol_desc(IsoImageWriter *writer)
     memcpy(vol.std_identifier, "CD001", 5);
     vol.vol_desc_version[0] = 1;
     memcpy(vol.boot_sys_id, "EL TORITO SPECIFICATION", 23);
-    iso_lsb(vol.boot_catalog, t->cat->sections[0].block, 4);
-
+    iso_lsb(vol.boot_catalog,
+            t->cat->sections[0].block - t->eff_partition_offset, 4);
     return iso_write(t, &vol, sizeof(struct ecma119_boot_rec_vol_desc));
 }
 
@@ -916,12 +1223,12 @@ int eltorito_writer_free_data(IsoImageWriter *writer)
 
 int eltorito_writer_create(Ecma119Image *target)
 {
-    int ret;
+    int ret, idx, outsource_efi = 0;
     IsoImageWriter *writer;
     IsoFile *bootimg;
     IsoFileSrc *src;
 
-    writer = malloc(sizeof(IsoImageWriter));
+    writer = calloc(1, sizeof(IsoImageWriter));
     if (writer == NULL) {
         return ISO_OUT_OF_MEM;
     }
@@ -947,20 +1254,49 @@ int eltorito_writer_create(Ecma119Image *target)
             return ret;
         }
     }
-    bootimg = target->catalog->image->image;
-    ret = iso_file_src_create(target, bootimg, &src);
-    if (ret < 0) {
-        return ret;
-    }
-    target->bootimg = src;
 
-    /* if we have selected to patch the image, it needs to be copied always */
-    if (target->catalog->image->isolinux_options & 0x01) {
-        src->prev_img = 0;
+    if (target->efi_boot_partition != NULL)
+        if (strcmp(target->efi_boot_partition, "--efi-boot-image") == 0)
+            outsource_efi = 1;
+    for (idx = 0; idx < target->catalog->num_bootimages; idx++) {
+        bootimg = target->catalog->bootimages[idx]->image;
+        ret = iso_file_src_create(target, bootimg, &src);
+        if (ret < 0) {
+            return ret;
+        }
+        target->bootsrc[idx] = src;
+
+        /* For patching an image, it needs to be copied always */
+        if (target->catalog->bootimages[idx]->isolinux_options & 0x01) {
+            src->no_write = 0;
+        }
+
+        /* If desired: Recognize first EFI boot image that will be newly
+           written, and mark it as claimed for being a partition.
+        */
+        if (outsource_efi &&
+            target->catalog->bootimages[idx]->platform_id == 0xef &&
+            src->no_write == 0) {
+           target->efi_boot_part_filesrc = src;
+           src->sections[0].block = 0xfffffffe;
+           ((IsoNode *) bootimg)->hidden |=
+                                   LIBISO_HIDE_ON_HFSPLUS | LIBISO_HIDE_ON_FAT;
+           outsource_efi = 0;
+        }
     }
 
     /* we need the bootable volume descriptor */
     target->curblock++;
+
+    if (outsource_efi) {
+        /* Disable EFI Boot partition and complain */
+        free(target->efi_boot_partition);
+        target->efi_boot_partition = NULL;
+        iso_msg_submit(target->image->id, ISO_BOOT_NO_EFI_ELTO, 0,
+"No newly added El Torito EFI boot image found for exposure as GPT partition");
+        return ISO_BOOT_NO_EFI_ELTO;
+    }
+
     return ISO_SUCCESS;
 }
 
